from module_common import SimulationTask
import re
import os
import subprocess
import time
import glob

__simulation__ = 'DemoSimulation'


class DemoSimulation(SimulationTask):

    def __init__(self, sim_id, name, full_dir, status, mode='daemon', t_min=0, t_max=0, restarts=None, logger=None):
        super(DemoSimulation, self).__init__(sim_id, name, full_dir, status, mode, t_min, t_max, restarts, logger)

    def sim_get_model_time(self):
        super(DemoSimulation, self).sim_get_model_time()
        orig_dir = os.getcwd()
        os.chdir(self.full_dir)
        if self.config.has_option('Simulation', 'Output_file'):
            output_file = self.config.get('Simulation', 'Output_file')
            regex = re.compile('\\d+')
            if os.path.isfile(output_file):
                last_line = subprocess.check_output(['tail', '-1', output_file])
                res = regex.findall(last_line)
                if len(res) > 0:
                    self.t = float(res[0])
        os.chdir(orig_dir)
        return self.t




"""
Implementation of utilities used by SiMon globally.
"""

import sys


config_file_template = """# Global config file for SiMon
[SiMon]

# The simulation data root directory
Root_dir: examples/demo_simulations

# The time interval for the SiMon daemon to check all the simulations (in seconds) [Default: 180]
Daemon_sleep_time: 180

# The number of simulations to be carried out simultaneously [Default: 2]
Max_concurrent_jobs: 2

# The maximum number of times a simulation will be restarted (a simulation is marked as ERROR when exceeding this limit) [Default: 2]
Max_restarts: 2

# Log level of the daemon: INFO/WARNING/ERROR/CRITICAL [default: INFO]
Log_level: INFO

# The time (in seconds) since the last modification of the output file, beyond which a simulation is considered stalled
Stall_time: 7200
"""


class Utilities(object):

    @staticmethod
    def progress_bar(val, val_max, val_min=0, prefix='', suffix='', bar_len=20):
        """
        Displays a progress bar in the simulation tree.
        :param val: current value
        :param val_max: maximum value
        :param val_min: minimum value
        :param prefix: marker for the completed part
        :param suffix: marker for the incomplete part
        :param bar_len: total length of the progress bar
        :return: a string representation of the progressbar
        """
        if val_max == 0:
            return ''
        else:
            skipped_len = int(round(bar_len * val_min) / float(val_max))
            filled_len = int(round(bar_len * (val - val_min) / float(val_max)))
            # percents = round(100.0 * count / float(total), 1)
            bar = '.' * skipped_len + '|' * filled_len + '.' * (bar_len - filled_len - skipped_len)
            # return '[%s] %s%s %s\r' % (bar, percents, '%', suffix)
            return '%s [%s] %s\r' % (prefix, bar, suffix)

    @staticmethod
    def highlighted_text(text, color=None, bold=False):
        colors = ['red', 'blue', 'cyan', 'green', 'reset']
        color_codes = ['\033[31m', '\033[34m', '\033[36m', '\033[32m', '\033[0m']
        color_codes_bold = ['\033[1;31m', '\033[1;34m', '\033[1;36m', '\033[0;32m', '\033[0;0m']

        if color not in colors:
            color = 'reset'
        if bold is False:
            return '%s%s%s' % (color_codes[colors.index(color)], text, color_codes[colors.index('reset')])
        else:
            return '%s%s%s' % (color_codes_bold[colors.index(color)], text, color_codes_bold[colors.index('reset')])

    @staticmethod
    def id_input(prompt):
        """
        Prompt to the user to input the simulation ID (in the interactive mode)
        """
        confirmed = False
        vec_index_selected = []
        while confirmed is False:
            response = Utilities.get_input(prompt)
            fragment = response.split(',')
            for token_i in fragment:
                if '-' in token_i:  # it is a range
                    limits = token_i.split('-')
                    if len(limits) == 2:
                        try:
                            if int(limits[0].strip()) < int(limits[1].strip()):
                                subrange = range(int(limits[0].strip()), int(limits[1].strip())+1)
                                for j in subrange:
                                    vec_index_selected.append(j)
                        except ValueError:
                            print('Invalid input. Please use only integer numbers.')
                            continue
                else:
                    try:
                        int(token_i.strip())  # test integer
                        vec_index_selected.append(token_i.strip())
                    except ValueError:
                        print('Invalid input %s. Please use only integer numbers.' % token_i.strip())
                        continue
            if Utilities.get_input('Your input is \n\t'+str(vec_index_selected)+', confirm? [Y/N] ').lower() == 'y':
                confirmed = True
                return map(int, vec_index_selected)
            else:
                vec_index_selected = []

    @staticmethod
    def get_input(prompt_msg):
        """
        This method makes use of the raw_input() method in Python2 and input() method in Python 3.
        """
        if sys.version_info[:2] <= (2, 7):
            return raw_input(prompt_msg)
        else:
            return input(prompt_msg)

    @staticmethod
    def generate_conf():
        try:
            target = open('SiMon.conf', 'w')
            target.write(config_file_template)
            target.close()
        except IOError:
            print("Unexpected error:", sys.exc_info()[0])



import os
import os.path
import sys
import time
import logging
import glob

from SiMon.utilities import Utilities
import numpy as np
try:
    import configparser as cp  # Python 3 only
except ImportError:
    import ConfigParser as cp  # Python 2 only
from fnmatch import fnmatch
from daemon import runner
from SiMon.module_common import SimulationTask

__simon_dir__ = os.path.dirname(os.path.abspath(__file__))
__user_shell_dir__ = os.getcwd()

class SiMon(object):
    """
    Main code of Simulation Monitor (SiMon).
    """

    def __init__(self, pidfile=None, stdin='/dev/tty', stdout='/dev/tty', stderr='/dev/tty',
                 mode='interactive', cwd=os.getcwd(), config_file='SiMon.conf'):
        """
        :param pidfile:
        """

        # Only needed in interactive mode
        conf_path = os.path.join(cwd, config_file)
        self.config = self.parse_config_file(conf_path)

        if self.config is None:
            print('Error: Configuration file SiMon.conf does not exist on the current path: %s' % cwd)
            if Utilities.get_input('Would you like to generate the default SiMon.conf file to the current directory? [Y/N] ').lower() == 'y':
                # shutil.copyfile(os.path.join(__simon_dir__, 'SiMon.conf'), os.path.join(cwd, 'SiMon.conf'))
                Utilities.generate_conf()
                print('SiMon.conf is now on the current directly. Please edit it accordingly and run ``simon [start|stop|interactive|i]``.')
            sys.exit(-1)
        else:
            try:
                cwd = self.config.get('SiMon', 'Root_dir')
            except cp.NoOptionError:
                print('Item Root_dir is missing in configuration file SiMon.conf. SiMon cannot start. Exiting...')
                sys.exit(-1)
        # make sure that cwd is the absolute path
        if not os.path.isabs(cwd):
            cwd = os.path.join(os.getcwd(), cwd)  # now cwd is the simulation data root directory
        if not os.path.isdir(cwd):
            if Utilities.get_input('Simulation root directory does not exist. '
                                   'Would you like to generate test simulations on the current directory? [Y/N] ').lower() == 'y':
                import SiMon.ic_generator_demo as ic_generator_demo
                ic_generator_demo.generate_ic(cwd)
                print('Demo simulations generated. Please start them with ``simon start``')
            else:
                print('Exiting...')
            sys.exit(-1)

        self.module_dict = self.register_modules()

        self.selected_inst = []  # A list of the IDs of selected simulation instances
        self.sim_inst_dict = dict()  # the container of all SimulationTask objects (ID to object mapping)
        self.sim_inst_parent_dict = dict()  # given the current path, find out the instance of the parent

        # TODO: create subclass instance according to the config file
        self.sim_tree = SimulationTask(0, 'root', cwd, SimulationTask.STATUS_NEW)
        self.stdin_path = stdin
        self.stdout_path = stdout
        self.stderr_path = stderr
        self.pidfile_path = pidfile
        self.pidfile_timeout = 5
        self.mode = mode
        self.cwd = cwd
        self.inst_id = 0
        self.logger = None
        self.max_concurrent_jobs = 2

        if self.config.has_option('SiMon', 'Max_concurrent_jobs'):
            self.max_concurrent_jobs = self.config.getint('SiMon', 'Max_concurrent_jobs')

        os.chdir(cwd)

    @staticmethod
    def parse_config_file(config_file):
        """
        Parse the configure file (SiMon.conf) for starting SiMon. The basic information of Simulation root directory
        must exist in the configure file before SiMon can start. A minimum configure file of SiMon looks like:

        ==============================================
        [SiMon]
        Root_dir: <the_root_dir_of_the_simulation_data>
        ==============================================

        :return: return 0 if succeed, -1 if failed (file not exist, and cannot be created). If the file does not exist
        but a new file with default values is created, the method returns 1.
        """
        conf = cp.ConfigParser()
        if os.path.isfile(config_file):
            conf.read(config_file)
            return conf
        else:
            return None

    @staticmethod
    def register_modules():
        """
        Register modules
        :return: A dict-like mapping between the name of the code and the filename of the module.
        """
        mod_dict = dict()
        module_candidates = glob.glob(os.path.join(__simon_dir__, 'module_*.py'))
        module_cwd = glob.glob(os.path.join(__user_shell_dir__, 'module_*.py'))  # load the modules also from cwd
        for m_cwd in module_cwd:
            module_candidates.append(m_cwd)
        for mod_name in module_candidates:
            sys.path.append(__simon_dir__)
            sys.path.append(os.getcwd())
            mod_name = os.path.basename(mod_name)
            mod = __import__(mod_name.split('.')[0])
            if hasattr(mod, '__simulation__'):
                # it is a valid SiMon module
                mod_dict[mod.__simulation__] = mod_name.split('.')[0]
        return mod_dict

    def traverse_simulation_dir_tree(self, pattern, base_dir, files):
        """
        Traverse the simulation file structure tree (Breadth-first search), until the leaf (i.e. no restart directory)
        or the simulation is not restartable (directory with the 'STOP' file).
        """
        for filename in sorted(files):
            if fnmatch(filename, pattern):
                if os.path.isdir(os.path.join(base_dir, filename)):
                    fullpath = os.path.join(base_dir, filename)
                    self.inst_id += 1
                    id = self.inst_id

                    # Try to determine the simulation code type by reading the config file
                    sim_config = self.parse_config_file(os.path.join(fullpath, 'SiMon.conf'))
                    sim_inst = None
                    if sim_config is not None:
                        try:
                            code_name = sim_config.get('Simulation', 'Code_name')
                            if code_name in self.module_dict:
                                sim_inst_mod = __import__(self.module_dict[code_name])
                                sim_inst = getattr(sim_inst_mod, code_name)(id, filename, fullpath,
                                                                            SimulationTask.STATUS_NEW,
                                                                            logger=self.logger)
                        except (cp.NoOptionError, cp.NoSectionError):
                            pass
                    else:
                        continue
                    if sim_inst is None:
                        continue
                    self.sim_inst_dict[id] = sim_inst
                    sim_inst.id = id
                    sim_inst.fulldir = fullpath
                    sim_inst.name = filename

                    # register child to the parent
                    self.sim_inst_parent_dict[base_dir].restarts.append(sim_inst)
                    sim_inst.level = self.sim_inst_parent_dict[base_dir].level + 1
                    # register the node itself in the parent tree
                    self.sim_inst_parent_dict[fullpath] = sim_inst
                    sim_inst.parent_id = self.sim_inst_parent_dict[base_dir].id

                    # Get simulation status
                    sim_inst.sim_get_status()

                    self.sim_inst_dict[sim_inst.parent_id].status = sim_inst.status

                    if (sim_inst.t > self.sim_inst_dict[sim_inst.parent_id].t and
                            not os.path.isfile(os.path.join(sim_inst.fulldir, 'ERROR'))) \
                            or sim_inst.status == SimulationTask.STATUS_RUN:
                        # nominate as restart candidate
                        self.sim_inst_dict[sim_inst.parent_id].cid = sim_inst.id
                        self.sim_inst_dict[sim_inst.parent_id].t_max_extended = sim_inst.t_max_extended

    def build_simulation_tree(self):
        """
        Generate the simulation tree data structure, so that a restarted simulation can trace back
        to its ancestor.

        :return: The method has no return. The result is stored in self.sim_tree.
        :type: None
        """
        os.chdir(self.cwd)
        self.sim_inst_dict = dict()

        self.sim_tree = SimulationTask(0, 'root', self.cwd, SimulationTask.STATUS_NEW)  # initially only the root node
        self.sim_inst_dict[0] = self.sim_tree  # map ID=0 to the root node
        self.sim_inst_parent_dict[self.cwd.strip()] = self.sim_tree  # map the current dir to be the sim tree root
        self.inst_id = 0

        for directory, dirnames, filenames in os.walk(self.cwd):
            self.traverse_simulation_dir_tree('*', directory, dirnames)

        # Synchronize the status tree (status propagation)
        update_needed = True
        max_iter = 0
        while update_needed and max_iter < 30:
            max_iter += 1
            inst_status_modified = False
            for i in self.sim_inst_dict:
                if i == 0:
                    continue
                inst = self.sim_inst_dict[i]
                if inst.status == SimulationTask.STATUS_RUN or inst.status == SimulationTask.STATUS_DONE:
                    if inst.parent_id > 0 and self.sim_inst_dict[inst.parent_id].status != inst.status:
                        # propagate the status of children (restarted simulation) to parents' status
                        self.sim_inst_dict[inst.parent_id].status = inst.status
                        inst_status_modified = True
            if inst_status_modified is True:
                update_needed = True
            else:
                update_needed = False
        return 0
        # print self.sim_tree

    def print_sim_status_overview(self, sim_id):
        """
        Output an overview of the simulation status in the terminal.

        :return: start and stop time
        :rtype: int
        """
        print(self.sim_inst_dict[sim_id])  # print the root node will cause the whole tree to be printed
        return self.sim_inst_dict[sim_id].t_min, self.sim_inst_dict[sim_id].t_max

    @staticmethod
    def print_help():
        print('Usage: python simon.py [start|stop|interactive|help]')
        print('\tTo show an overview of job status and quit: python simon.py (no arguments)')
        print('\tstart: start the daemon')
        print('\tstop: stop the daemon')
        print('\tinteractive/i/-i: run in interactive mode (no daemon)')
        print('\thelp: print this help message')

    @staticmethod
    def print_task_selector():
        """
        Prompt a menu to allow the user to select a task.

        :return: current selected task symbol.
        """
        opt = ''
        while opt.lower() not in ['l', 's', 'n', 'r', 'c', 'x', 't', 'd', 'k', 'b', 'p', 'q']:
            sys.stdout.write('\n=======================================\n')
            sys.stdout.write('\tList Instances (L), \n\tSelect Instance (S), '
                             '\n\tNew Run (N), \n\tRestart (R), \n\tCheck status (C), '
                             '\n\tStop Simulation (T), \n\tDelete Instance (D), \n\tKill Instance (K), '
                             '\n\tBackup Restart File (B), \n\tPost Processing (P), \n\tUNIX Shell (X), '
                             '\n\tQuit (Q): \n')
            opt = Utilities.get_input('\nPlease choose an action to continue: ').lower()

        return opt

    def task_handler(self, opt):
        """
        Handles the task selection input from the user (in the interactive mode).

        :param opt: The option from user input.
        """

        if opt == 'q':  # quit interactive mode
            sys.exit(0)
        if opt == 'l':  # list all simulations
            self.build_simulation_tree()
            self.print_sim_status_overview(0)
        if opt in ['s', 'n', 'r', 'c', 'x', 't', 'd', 'k', 'b', 'p']:
            if self.mode == 'interactive':
                if self.selected_inst is None or len(self.selected_inst) == 0 or opt == 's':
                    self.selected_inst = Utilities.id_input('Please specify a list of IDs (seperated by comma): ')
                    sys.stdout.write('Instances ' + str(self.selected_inst) + ' selected.\n')

        # TODO: use message? to rewrite this part in a smarter way
        if opt == 'n':  # start new simulations
            for sid in self.selected_inst:
                if sid in self.sim_inst_dict:
                    self.sim_inst_dict[sid].sim_start()
                    # reset the selection list
                    self.selected_inst = []
                else:
                    print('The selected simulation with ID = %d does not exist. Simulation not started.\n' % sid)
        if opt == 'r':  # restart simulations
            for sid in self.selected_inst:
                if sid in self.sim_inst_dict:
                    self.sim_inst_dict[sid].sim_restart()
                    # reset the selection list
                    self.selected_inst = []
                else:
                    print('The selected simulation with ID = %d does not exist. Simulation not restarted.\n' % sid)
        if opt == 'c':  # check the recent or current status of the simulation and print it
            for sid in self.selected_inst:
                if sid in self.sim_inst_dict:
                    self.sim_inst_dict[sid].sim_collect_recent_output_message()
                else:
                    print('The selected simulation with ID = %d does not exist. Simulation not restarted.\n' % sid)
        if opt == 'x':  # execute an UNIX shell command in the simulation directory
            print('Executing an UNIX shell command in the selected simulations.')
            shell_command = Utilities.get_input('CMD>> ')
            for sid in self.selected_inst:
                if sid in self.sim_inst_dict:
                    self.sim_inst_dict[sid].sim_shell_exec(shell_command=shell_command)
                else:
                    print('The selected simulation with ID = %d does not exist. Cannot execute command.\n' % sid)
        if opt == 't':  # soft-stop the simulation in the ways that supported by the code
            for sid in self.selected_inst:
                if sid in self.sim_inst_dict:
                    self.sim_inst_dict[sid].sim_stop()
                else:
                    print('The selected simulation with ID = %d does not exist. Simulation not stopped.\n' % sid)
        if opt == 'd':  # delete the simulation tree and all its data
            for sid in self.selected_inst:
                if sid in self.sim_inst_dict:
                    self.sim_inst_dict[sid].sim_delete()
                    # reset the selection list
                    self.selected_inst = []
                else:
                    print('The selected simulation with ID = %d does not exist. Cannot delete simulation.\n' % sid)
        if opt == 'k':  # kill the UNIX process associate with a simulation task
            for sid in self.selected_inst:
                if sid in self.sim_inst_dict:
                    self.sim_inst_dict[sid].sim_kill()
                    # reset the selection list
                    self.selected_inst = []
                else:
                    print('The selected simulation with ID = %d does not exist. Cannot kill simulation.\n' % sid)
        if opt == 'b':  # backup the simulation checkpoint files (for restarting purpose in the future)
            for sid in self.selected_inst:
                if sid in self.sim_inst_dict:
                    self.sim_inst_dict[sid].sim_backup_checkpoint()
                else:
                    print('The selected simulation with ID = %d does not exist. Cannot backup checkpoint.\n' % sid)
        if opt == 'p':  # perform (post)-processing (usually after the simulation is done)
            for sid in self.selected_inst:
                if sid in self.sim_inst_dict:
                    self.sim_inst_dict[sid].sim_finalize()
                    # reset the selection list
                    self.selected_inst = []
                else:
                    print('The selected simulation with ID = %d does not exist. Cannot perform postprocessing.\n' % sid)

    def auto_scheduler(self):
        """
        The automatic decision maker for the daemon.

        The daemon invokes this method at a fixed period of time. This method checks the
        status of all simulations by traversing to all simulation directories and parsing the
        output files. It subsequently deals with the simulation instance according to the informtion
        gathered.
        """
        os.chdir(self.cwd)
        self.build_simulation_tree()
        schedule_list = []
        # Sort jobs according to priority (niceness)
        sim_niceness_vec = []

        # check how many simulations are running
        concurrent_jobs = 0
        for i in self.sim_inst_dict.keys():
            inst = self.sim_inst_dict[i]
            sim_niceness_vec.append(inst.niceness)
            inst.sim_get_status()  # update its status
            # test if the process is running
            if inst.status == SimulationTask.STATUS_RUN and inst.cid == -1:
                concurrent_jobs += 1

        index_niceness_sorted = np.argsort(sim_niceness_vec)
        for ind in index_niceness_sorted:
            if self.sim_inst_dict[ind].status != SimulationTask.STATUS_DONE and self.sim_inst_dict[ind].id > 0:
                schedule_list.append(self.sim_inst_dict[ind])
                print(self.sim_inst_dict[ind].name)

        for sim in schedule_list:
            if sim.id == 0:  # the root group, skip
                continue
            sim.sim_get_status()  # update its status
            print('Checking instance #%d ==> %s [%s]' % (sim.id, sim.name, sim.status))
            if sim.status == SimulationTask.STATUS_RUN:
                sim.sim_backup_checkpoint()
            elif sim.status == SimulationTask.STATUS_STALL:
                sim.sim_kill()
                self.build_simulation_tree()
            elif sim.status == SimulationTask.STATUS_STOP and sim.level == 1:
                self.logger.warning('STOP detected: '+sim.fulldir)
                # check if there is available slot to restart the simulation
                if concurrent_jobs < self.max_concurrent_jobs and sim.level == 1:
                    # search only top level instance to find the restart candidate
                    # build restart path
                    current_inst = sim
                    # restart the simulation instance at the leaf node
                    while current_inst.cid != -1:
                        current_inst = self.sim_inst_dict[current_inst.cid]
                    print('RESTART: #%d ==> %s' % (current_inst.id, current_inst.fulldir))
                    self.logger.info('RESTART: #%d ==> %s' % (current_inst.id, current_inst.fulldir))
                    current_inst.sim_restart()
                    concurrent_jobs += 1
            elif sim.status == SimulationTask.STATUS_NEW:
                # check if there is available slot to start the simulation
                if concurrent_jobs < self.max_concurrent_jobs:
                    # Start new run
                    sim.sim_start()
                    concurrent_jobs += 1
        self.logger.info('SiMon routine checking completed. Machine load: %d/%d' % (concurrent_jobs,
                                                                                    self.max_concurrent_jobs))

    def run(self):
        """
        The entry point of this script if it is run with the daemon.
        """
        os.chdir(self.cwd)
        self.build_simulation_tree()
        while True:
            # print('[%s] Auto scheduled' % datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S'))
            self.auto_scheduler()
            sys.stdout.flush()
            sys.stderr.flush()
            if self.config.has_option('SiMon', 'daemon_sleep_time'):
                time.sleep(self.config.getfloat('SiMon', 'daemon_sleep_time'))
            else:
                time.sleep(180)

    def interactive_mode(self, autoquit=False):
        """
        Run SiMon in the interactive mode. In this mode, the user can see an overview of the simulation status from the
        terminal, and control the simulations accordingly.
        :return:
        """
        print(os.getcwd())
        os.chdir(self.cwd)
        self.build_simulation_tree()
        self.print_sim_status_overview(0)
        choice = ''
        if autoquit is False:
            while choice != 'q':
                choice = SiMon.print_task_selector()
                self.task_handler(choice)

    @staticmethod
    def daemon_mode(simon_dir):
        """
        Run SiMon in the daemon mode.

        In this mode, SiMon will behave as a daemon process. It will scan all simulations periodically, and take measures
        if necessary.
        :return:
        """
        app = SiMon(pidfile=os.path.join(simon_dir, 'SiMon_daemon.pid'),
                    stdout=os.path.join(simon_dir, 'SiMon.out.txt'),
                    stderr=os.path.join(simon_dir, 'SiMon.err.txt'),
                    cwd=simon_dir,
                    mode='daemon')
        # log system
        app.logger = logging.getLogger("DaemonLog")
        if app.config.has_option('SiMon', 'Log_level'):
            log_level = app.config.get('SiMon', 'Log_level')
            if log_level == 'INFO':
                app.logger.setLevel(logging.INFO)
            elif log_level == 'WARNING':
                app.logger.setLevel(logging.WARNING)
            elif log_level == 'ERROR':
                app.logger.setLevel(logging.ERROR)
            elif log_level == 'CRITICAL':
                app.logger.setLevel(logging.CRITICAL)
            else:
                app.logger.setLevel(logging.INFO)
        formatter = logging.Formatter("%(asctime)s - [%(levelname)s] - %(name)s - %(message)s")
        handler = logging.FileHandler(os.path.join(simon_dir, 'SiMon.log'))
        handler.setFormatter(formatter)
        app.logger.addHandler(handler)
        # initialize the daemon runner
        app.logger.info('Starting SiMon daemon at log level %s' % log_level)
        daemon_runner = runner.DaemonRunner(app)
        # This ensures that the logger file handle does not get closed during daemonization
        daemon_runner.daemon_context.files_preserve = [handler.stream]
        daemon_runner.do_action()  # fixed time period of calling run()


def main():
    # execute only if run as a script
    if len(sys.argv) == 1:
        print('Running SiMon in the interactive mode...')
        s = SiMon()
        s.interactive_mode(autoquit=True)
    elif len(sys.argv) > 1:
        if sys.argv[1] in ['start', 'stop', 'restart']:
            if sys.argv[1] == 'start':
                # test if the daemon is already started
                if os.path.isfile('SiMon_daemon.pid'):
                    try:
                        f_pid = open('SiMon_daemon.pid')
                        simon_pid = int(f_pid.readline())
                        os.kill(simon_pid, 0)  # test whether the process exists, does not kill the process
                        print('Error: the SiMon daemon is already running with process ID: %d' % simon_pid)
                        print('Please make sure that you stop the daemon before starting it. Exiting...')
                        sys.exit(-1)
                    except (ValueError, OSError):
                        pass
            # The python-daemon library will handle the start/stop/restart arguments by itself
            try:
                SiMon.daemon_mode(os.getcwd())
            except runner.DaemonRunnerStopFailureError:
                print('Error: the SiMon daemon is not running. There is no need to stop it.')
                sys.exit(-1)
        elif sys.argv[1] in ['interactive', 'i', '-i']:
            s = SiMon()
            s.interactive_mode()
        else:
            print(sys.argv[1])
            SiMon.print_help()
            sys.exit(0)

if __name__ == "__main__":
    main()



import os
import numpy as np
from SiMon.ic_generator import InitialConditionGenerator


def generate_ic(output_basedir=os.getcwd()):
    # parameter space
    a_vec = [1.0, 2.0, 3.0]
    o_vec = [3.5, 7.5, 10.5, 16.5]
    t_end = 30.0

    # templates
    code_name = 'DemoSimulation'
    executable_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'demo_simulation_code.py')
    start_cmd_template = 'python -u %s -a %f -o %f -t %f -p %f 1>%s 2>%s'
    restart_cmd_template = 'cp ../%s . ; python -u %s -a %f -o %f -t %f -p %f 1>%s 2>%s'
    stop_cmd = 'touch STOP'
    output_dir_template = 'demo_sim_t_end=%g_a=%g_e=%g'

    # IC generator
    ic = InitialConditionGenerator(conf_file='SiMon.conf')
    ic.parse_config_file()

    # generate the IC parameter space in the loops
    for a in a_vec:
        for o in o_vec:
            p_crash = 0.01 * np.random.rand()
            start_cmd = start_cmd_template % (executable_path, a, o, t_end, p_crash, 'output.txt', 'error.txt')
            restart_cmd = restart_cmd_template % ('restart.txt', executable_path, a, o, t_end, p_crash, 'output.txt', 'error.txt')
            output_dir = output_dir_template % (t_end, a, o)
            ic.generate_simulation_ic(code_name, t_end, output_dir, start_cmd,
                                      input_file='input.txt', output_file='output.txt', error_file='error.txt',
                                      restart_cmd=restart_cmd, stop_cmd=stop_cmd)

if __name__ == "__main__":
    generate_ic()



import os
import sys
import time
import numpy as np
from optparse import OptionParser

parser = OptionParser()
parser.add_option('-s', '--t_start', action='store', type='float', dest='t_start',
                  help='The termination time of the pseudo simulation')
parser.add_option('-t', '--t_end', action='store', type='float', dest='t_end',
                  help='The termination time of the pseudo simulation')
parser.add_option('-d', '--dt', action='store', type='float', dest='dt',
                  help='The time step of the pseudo simulation')
parser.add_option('-f', '--output_file', action='store', type='string', dest='out_file',
                  help='The file name of the output file')
parser.add_option('-o', '--omega', action='store', type='float', dest='o',
                  help='frequency')
parser.add_option('-a', '--amplitude', action='store', type='float', dest='a',
                  help='The semi-major axis')
parser.add_option('-p', '--crash_probability', action='store', type='float', dest='p_crash',
                  help='The probability for the code to crash [0-1]')

t_start = 0.0
t_end = 10.0
dt = 0.02
out_file = 'output.txt'
restart_file = 'restart.txt'
restart_freq = 50
a = 1.0
omega = 1.0
p_crash = 0.001

(options, args) = parser.parse_args()
if options.t_start is not None:
    t_start = options.t_start
if options.t_end is not None:
    t_end = options.t_end
if options.dt is not None:
    dt = options.dt
if options.out_file is not None:
    out_file = options.out_file
if options.a is not None:
    a = options.a
if options.o is not None:
    omega = options.o
if options.p_crash is not None:
    p_crash = options.p_crash

if os.path.isfile(restart_file):
    f_restart = open(restart_file, 'r')
    t_start = float(f_restart.read())
    f_restart.close()

xx = np.arange(t_start, t_end+dt, dt)
yy = a*np.sin(omega*xx) + 0.05*a*np.random.rand(len(xx))

out_f = open(out_file, 'wb')

for ind in range(len(xx)):
    print(xx[ind])
    out_f.write('%f, %f\n' % (xx[ind], yy[ind]))
    if (ind != 0 and ind % restart_freq == 0) or (ind == len(xx) - 1):
        f_restart = open(restart_file, 'w')
        f_restart.write('%f' % xx[ind])
        f_restart.close()
        # if a STOP file is present in the current directory, stop the demo simulation
        if os.path.isfile('STOP'):
            sys.exit(0)
    time.sleep(0.05)
    if np.random.rand() < p_crash:
        # print('Code crashed!')
        sys.exit(-1)

out_f.close()



import datetime
import abc
import glob
import os
import subprocess
import signal
import time
import sys
import re
import shutil
from SiMon.utilities import Utilities

try:
    import configparser as cp  # Python 3 only
except ImportError:
    import ConfigParser as cp  # Python 2 only


class SimulationTask(object):
    """
    A simulation task is a single simulation which the user requests to finish.
    It is associated with 1) a set of initial conditions specified in the input file,
    2) a (bash) script to start up the code, 3) the status of the simulation (RUN/STOP/model time,
    start timestamp, last output timestamp, parent simulation ID if it is a restart, etc),
    and 4) the ending time of the simulation.

    Notes
    -----
    Traverse the directory structure:
    A hierarchical directory structure may form for a simulation that has been started for multiple times.
    For instance, a simulation is running on the directory '/sim1'. It crashes at T=120. So SiMon
    restarts it by creating a restart directory '/sim1/restart1'. 'restart1' runs until T=200, and then
    again crashes. So SiMon creates '/sim1/restart1/restart1' in attempt to start from T=200.

    """
    # Constants
    STATUS_NEW = 0x0  # newly initialized simulation
    STATUS_STOP = 0x1  # crashed simulation
    STATUS_RUN = 0x2  # the simulation is running
    STATUS_STALL = 0x3  # the simulation is running, but stalled
    STATUS_DONE = 0x4  # the simulation has finished
    STATUS_ERROR = 0x5  # the simulation is crashed and not possible to be restarted

    STATUS_LABEL = ['NEW', 'STOP', 'RUN', 'STALL', 'DONE', 'ERROR']

    __metaclass__ = abc.ABCMeta

    def __init__(self, sim_id, name, full_dir, status, mode='daemon', t_min=0, t_max=0, restarts=None, logger=None):
        """
        :param sim_id:

        :param name: Usually the name of the simulation directory.
        :type name: basestring

        :param full_dir: The full path of the simulation directory.

        :param status: RUN, STOP, RESTARTED

        :param t_min: default as 0
        :param t_max: default as 0
        :param restarts: default as None
        :return:
        """
        self.id = sim_id
        self.name = name
        self.full_dir = full_dir
        self.status = status
        self.logger = logger
        self.error_type = ''
        self.config_file = 'SiMon.conf'  # the file name of the config file to be placed in each simulation directory
        self.config = None
        self.t = 0  # the current model time
        self.t_min = t_min  # minimum time for the simulation to start
        self.t_max = t_max  # maximum time marking the completion of the simulation
        self.t_max_extended = t_max  # extended t_max by restarts
        self.mtime = 0  # timestamp of the last modification of the simulation output files
        self.ctime = 0  # timestamp of the creation of the simulation output files

        # the candidate instance ID to restart in case crashes
        # (-1: no candidate, restart from itself;)
        # (>0: restart from the candidate. If the candidate cannot restart, try siblings)
        self.cid = -1

        self.level = 0
        self.parent_id = -1
        self.mode = mode
        self.niceness = 0  # Priority, same as UNIX (-20 ~ 19, the lower ==> higher priority)
        self.maximum_number_of_checkpoints = 20
        if restarts is None:
            self.restarts = list()
        else:
            self.restarts = restarts  # children
        self.parse_config_file()
        self.sim_get_status()

    def __repr__(self, level=0):
        if level == 0:
            ret = '[%s] %s\n' % (SimulationTask.STATUS_LABEL[self.status], self.full_dir)
        else:
            # placeholder_dash = "|" + '-' * (level * 4)
            placeholder_dash = "| "
            placeholder_space = ' ' * ((level-1) * 4)
            mtime_str = datetime.datetime.fromtimestamp(self.mtime).strftime('%m-%d %H:%M')

            prefix = 'T: %g >> %g >> %g' % (int(self.t_min), int(self.t), int(self.t_max))
            suffix = mtime_str
            progress_bar = Utilities.progress_bar(self.t, self.t_max, self.t_min, prefix=prefix, suffix=suffix)

            info = "%s    \t%s\t" % (Utilities.highlighted_text(str(self.name), 'cyan', bold=True), progress_bar)

            ret = "[%s]\t%s%d%s%s\n" % (SimulationTask.STATUS_LABEL[self.status], placeholder_space, self.id, placeholder_dash, info)
            # ret = "    "*level+str(self.id)+repr(self.name)+"\n"
        for child in self.restarts:
            ret += child.__repr__(level + 1)
        return ret

    def parse_config_file(self):
        """
        Parse the configure file (SiMon.conf) for the simulation. If the file does not exist, a new file with default
        values will be created.

        :return: return 0 if succeed, -1 if failed (file not exist, and cannot be created). If the file does not exist
        but a new file with default values is created, the method returns 1.
        """
        conf_fn = os.path.join(self.full_dir, self.config_file)
        conf = cp.ConfigParser()
        if os.path.isfile(conf_fn):
            conf.read(conf_fn)
            self.config = conf
            # synchronize config options to attributes
            if self.config.has_option('Simulation', 'T_end'):
                self.t_max = self.config.getfloat('Simulation', 'T_end')
            if self.config.has_option('Simulation', 'T_start'):
                self.t_min = self.config.getfloat('Simulation', 'T_start')
            if self.config.has_option('Simulation', 'Niceness'):
                self.niceness = self.config.getint('Simulation', 'Niceness')
            if self.config.has_option('Simulation', 'Maximum_n_checkpoints'):
                self.maximum_number_of_checkpoints = self.config.getint('Simulation', 'Maximum_n_checkpoints')

        else:
            if self.id > 0:
                msg = 'WARNING: Simulation configuration file not exists! ' \
                      'Creating default configuration as SiMon.conf.\n'
                if self.logger is not None:
                    self.logger.warning(msg)
            # TODO: write default config file
        return 0

    def sim_start(self):
        """
        Start a new simulation.

        :return: Return 0 if succeed, -1 if failed. If the simulation is already started, then it will do nothing
        but return 1.
        """
        start_script_template = '%s & echo $!>.process.pid'
        orig_dir = os.getcwd()
        os.chdir(self.full_dir)

        # Test if the process is running accoding to the .process.pid file
        if os.path.isfile('.process.pid'):
            # if the PID file exists, try to read the process ID
            f_pid = open('.process.pid', 'r')
            pid = int(f_pid.readline().strip())
            f_pid.close()
            if pid > 0:
                try:
                    os.kill(pid, 0)
                    return 1  # if no exception, the process is already running
                except (ValueError, OSError):
                    pass  # process not started yet
        # If the process is not started yet, then start it in a normal way
        if self.config.has_option('Simulation', 'Start_command'):
            start_cmd = self.config.get('Simulation', 'Start_command')
            # self.proc = subprocess.Popen(start_cmd, shell=True)
            os.system(start_script_template % start_cmd)
            # sleep for a little while to make sure that the pid file exist
            time.sleep(0.5)
            fpid = open('.process.pid', 'r')
            pid = int(fpid.readline())
            fpid.close()
            self.config.set('Simulation', 'PID', str(pid))
            self.config.set('Simulation', 'Timestamp_started', str(time.time()))
            self.config.write(open(self.config_file, 'w'))
            if self.logger is not None:
                msg = 'Simulation %s started, PID = %d' % (self.name, pid)
                self.logger.info(msg)
        else:
            return -1
        os.chdir(orig_dir)
        return 0

    def sim_restart(self):
        """
        Restart the simulation.

        :return: Return 0 if succeed, -1 if failed. If the simulation is already running, then restart is not
        necessary, the method will do nothing but return 1. If the simulation is marked as 'STOP' or 'ERROR', then
        return 2 and do nothing.
        """
        if os.path.isfile('STOP') or os.path.isfile('ERROR'):
            print('Restart skipped due to the existence of the STOP file or ERROR file.')
            return 2
        # Test if the process is running
        restart_script_template = '%s & echo $!>.process.pid'
        orig_dir = os.getcwd()
        os.chdir(self.full_dir)
        print('The full dir is %s' % self.full_dir)
        print('restarting simulation: %s' % self.full_dir)
        if self.logger is not None:
            self.logger.info('Restarting simulation: %s' % self.full_dir)
        # Test if the process is running accoding to the .process.pid file
        if os.path.isfile('.process.pid'):
            # if the PID file exists, try to read the process ID
            f_pid = open('.process.pid', 'r')
            pid = int(f_pid.readline().strip())
            f_pid.close()
            if pid > 0:
                try:
                    os.kill(pid, 0)
                    return 1  # if no exception, the process is already running
                except (ValueError, OSError):
                    # process not started yet
                    # check how many times the simulation has been restarted
                    restarts = glob.glob('restart*/')
                    n_restarts = len(restarts)
                    print(n_restarts, self.config.getint('Simulation', 'Max_restarts'))
                    # check whether it exceeds the maximum times of restarts specified in the per-sim config file
                    if self.config.has_option('Simulation', 'Max_restarts'):
                        if n_restarts > self.config.getint('Simulation', 'Max_restarts'):
                            # if exceed, create an empty file called 'ERROR'
                            f_error = open('ERROR', 'w')
                            f_error.close()
                            msg = 'Simulation %s has been restarted too many times. Further restart skipped...' % self.full_dir
                            print(msg)
                            if self.logger is not None:
                                self.logger.error(msg)
                            return -2
                    else:
                        # if the config entry Max_restarts does not exist in the config file, there is no restart limit
                        pass
                    # now try to restart the simulation
                    if self.config.has_option('Simulation', 'Restart_command'):
                        restart_cmd = self.config.get('Simulation', 'Restart_command')
                        if restart_cmd is not '' and restart_cmd.strip() is not 'None':
                            msg = 'Restarting simulation: %s' % self.full_dir
                            print(msg)
                            if self.logger is not None:
                                self.logger.info(msg)
                            # create a restart dir
                            restart_dir = 'restart%d' % (n_restarts + 1)
                            os.mkdir(restart_dir)
                            os.chdir(restart_dir)
                            os.system(restart_script_template % restart_cmd)
                            # sleep for a little while to make sure that the pid file exist
                            time.sleep(0.5)
                            fpid = open('.process.pid', 'r')
                            pid = int(fpid.readline())
                            fpid.close()
                            self.config.set('Simulation', 'PID', str(pid))
                            self.config.set('Simulation', 'Timestamp_started', str(time.time()))
                            self.config.write(open(self.config_file, 'w'))
                        else:
                            msg = '%s: unable to restart because the restart command is not properly configured.' % self.name
                            print(msg)
                            if self.logger is not None:
                                self.logger.error(msg)
                            return -1
                    else:
                        msg = '%s: unable to restart because the restart command is not configured.' % self.name
                        print(msg)
                        if self.logger is not None:
                            self.logger.error(msg)
                        return -1
        os.chdir(orig_dir)
        return 0

    def sim_get_model_time(self):
        """
        Get the model time of the simulation.

        Because different codes have different output formats, there is no generic way to obtain the model
        time. The user is required to implement this method to properly obtain the time.

        :return: the current model time
        """
        orig_dir = os.getcwd()
        os.chdir(self.full_dir)
        if self.config.has_option('Simulation', 'Output_file'):
            output_file = self.config.get('Simulation', 'Output_file')
            regex = re.compile('\\d+')
            if os.path.isfile(output_file):
                last_line = subprocess.check_output(['tail', '-1', output_file]).decode('utf-8')
                res = regex.findall(last_line)
                if len(res) > 0:
                    self.t = float(res[0])
        os.chdir(orig_dir)
        return self.t

    def sim_get_model_start_time(self):
        """
        Get the t_min value of the current model
        :return: The starting time of the current model. By default, t_min = 0. But if a model is restarted, then
        it is possible that t_min equals the time of the restartable snapshot.
        The user is required to implement this method properly.
        """
        if self.t_min != 0.0:
            return self.t_min
        else:
            return 0.0

    def sim_get_model_termination_time(self):
        """

        Get the time t_max by which the model is considered finished. Normally this value is obtained by parsing the
        config file, but the user may override it.

        :return: The termination time
        """
        return self.t_max

    def sim_get_status(self):
        """
        Get the current status of the simulation. Update the config file if necessary.

        :return: The code of the current simulation status.
        """
        if self.config is None:
            return 0
        orig_dir = os.getcwd()
        os.chdir(self.full_dir)
        self.t = self.sim_get_model_time()
        self.t_min = self.sim_get_model_start_time()

        # Check the last output time from either the output file or the error file
        output_file = ''
        if self.config.has_option('Simulation', 'Output_file'):
            output_file = self.config.get('Simulation', 'Output_file')
            if os.path.isfile(output_file):
                self.mtime = os.stat(output_file).st_mtime
        elif self.config.has_option('Simulation', 'Error_file'):
            error_file = self.config.get('Simulation', 'Error_file')
            if os.path.isfile(error_file_file):
                self.mtime = os.stat(error_file).st_mtime

        # Get the starting time of the simulation
        if self.config.has_option('Simulation', 'Timestamp_started'):
            self.ctime = self.config.getfloat('Simulation', 'Timestamp_started')

        # Determine whether the simulation is running using the process ID
        if os.path.isfile('.process.pid'):
            # if the PID file exists, try to read the process ID
            f_pid = open('.process.pid', 'r')
            pid = int(f_pid.readline().strip())
            f_pid.close()
            if pid == 0:
                if self.mtime == 0:
                    self.status = SimulationTask.STATUS_NEW
            else:
                try:
                    os.kill(pid, 0)  # This just checks if the process is running. It doesn't kill the process
                    # It is running. Check if stalled.
                    # The default value is large to prevent a slow simulation to be mistakenly killed
                    stall_time = 6.e6  # after 6.e6 seconds if the code doesn't advance, it is considered stalled
                    if self.config.has_option('Simulation', 'Stall_time'):
                        # Allow overriding the stall time using the per-simulation config file
                        stall_time = self.config.getfloat('Simulation', 'Stall_time')
                    if time.time() - self.mtime > stall_time:
                        self.status = SimulationTask.STATUS_STALL
                        if self.logger is not None:
                            mtime_str = datetime.datetime.fromtimestamp(self.mtime).strftime('%m-%d %H:%M')
                            msg = 'job %s is running [PID=%d], but no update in its output file (%s) since %s. ' \
                                  'The stall time of this task is %s sec. ' \
                                  'Marked as STALL' % (self.name, pid, output_file, mtime_str, stall_time)
                            print(msg)
                            self.logger.info(msg)
                    else:
                        self.status = SimulationTask.STATUS_RUN
                except (OSError, ValueError) as e:
                    # The process is not running, check if stopped or done
                    if self.t >= self.t_max or self.status == SimulationTask.STATUS_DONE:
                        self.status = SimulationTask.STATUS_DONE
                    else:
                        if self.ctime == 0.0:
                            self.status = SimulationTask.STATUS_NEW
                        elif os.path.isfile('ERROR'):
                            self.status = SimulationTask.STATUS_ERROR
                        else:
                            self.status = SimulationTask.STATUS_STOP
        os.chdir(orig_dir)
        return self.status

    def sim_kill(self):
        """
        Forcibly kill (i.e. terminate) the current simulation. Practically, this method terminates the process of
        the simulation code and sets the simulation status to STOP.

        :return: Return 0 if succeed, -1 if failed. If the simulation is not running, then it cannot be killed, causing
        the method to do nothing but return 1.
        """
        orig_dir = os.getcwd()
        os.chdir(self.full_dir)
        # Find the process by PID
        if os.path.isfile('.process.pid'):
            # if the PID file exists, try to read the process ID
            f_pid = open('.process.pid', 'r')
            pid = int(f_pid.readline().strip())
            f_pid.close()
            try:
                os.kill(pid, signal.SIGKILL)
                msg = 'Simulation %s (PID: %d) killed.' % (self.name, pid)
                print(msg)
                if self.logger is not None:
                    self.logger.info(msg)
            except OSError as err:
                msg = '%s: Cannot kill the job `%s` with PID = %d\n' % (str(err),  self.name, pid)
                print(msg)
                if self.logger is not None:
                    self.logger.error(msg)
        os.chdir(orig_dir)
        return 0

    def sim_stop(self):
        """
        Submit a request to the simulation code, in attempt to stop the simulation before it finishes the originally
        planned time checkpoint. This method will ask the code to stop the simulation by itself (if supported), rather
        than forcibly killing the simulation process.

        :return: Return 0 if succeed, -1 if failed. If the simulation is not running, then it cannot be stopped, causing
        the method to do nothing but return 1.
        """
        # Create an empty file called 'STOP'. The integrator that detects this file will (hopefully) stop the
        # integration.
        orig_dir = os.getcwd()
        os.chdir(self.full_dir)
        stop_file = open(os.path.join(self.full_dir, 'STOP'), 'w')
        stop_file.close()
        msg = 'A stop request has been sent to simulation %s' % self.name
        print(msg)
        if self.logger is not None:
            self.logger.info(msg)
        os.chdir(orig_dir)
        return 0

    def sim_backup_checkpoint(self):
        """
        Back up a snapshot of the latest restart files or simulation snapshot. In case of code crash, the backup files
        can be used for restarting.

        :return: Return 0 if succeed, -1 if failed. If the existing simulation snapshot is already the latest version,
        backup is not necessary, causing the method to do nothing but return 1.
        """
        # Try to get the restartable checkpoint file name from the config file
        orig_dir = os.getcwd()
        os.chdir(self.full_dir)
        if self.config.has_option('Simulation', 'Restart_file'):
            restart_fn = self.config.get('Simulation', 'Restart_file')
            ts = time.time()  # get the timestamp as part of the backup restart file name
            backup_restart_fn = 'restart.tmp.%d' % int(ts)
            if os.path.isfile(os.path.join(self.full_dir, restart_fn)):
                shutil.copyfile(os.path.join(self.full_dir, restart_fn), os.path.join(self.full_dir, backup_restart_fn))
                msg = 'Restart file has been backup as ' + backup_restart_fn
                print(msg)
                if self.logger is not None:
                    self.logger.info(msg)
                # delete the oldest backup if there is a limit of maximum number of backup files
                backup_file_list = sorted(glob.glob(os.path.join(self.full_dir, 'restart.tmp.*')))
                if 0 < self.maximum_number_of_checkpoints < len(backup_file_list):
                    for backup_restart_del_fn in backup_file_list[:-abs(self.maximum_number_of_checkpoints)]:
                        os.remove(os.path.join(self.full_dir, backup_restart_del_fn))
        else:
            # Without knowing the name of the restartable snapshot, SiMon will not be able to backup
            if self.logger is not None:
                self.logger.info('SiMon does not know how to backup the current simulation %s' % self.name)
            return -1
        os.chdir(orig_dir)
        return 0

    def sim_delete(self):
        """
        Delete the simulation data (including restarted simulation data).

        :return: Return 0 if succeed, -1 if failed. A simulation cannot be deleted if it is currently running. In this
        case, this method does nothing but just return 1.
        """
        if self.mode == 'interactive':
            confirm = Utilities.get_input('Are you sure you would like to delete the instance '
                                          '#%d and its sub-instances? [Y/N] ' % self.id).lower()
            if confirm != 'y':
                return 1
        else:
            # TODO: code will not goes here because no functions in daemon mode will call inst_delete
            shutil.rmtree(self.full_dir)
            return 0
        return 0

    def sim_shell_exec(self, shell_command=None):
        """
        Execute a shell command under the data directory of the simulation.

        :param shell_command: the shell command to execute

        :return: Return 0 if succeed, -1 if failed.
        """
        if shell_command is None:
            shell_command = Utilities.get_input('CMD>> ')
        sys.stdout.write('========== Command on #%d ==> %s (PWD=%s) ==========\n'
                         % (self.id, self.full_dir, self.full_dir))
        original_dir = os.getcwd()
        os.chdir(self.full_dir)
        os.system(shell_command)
        sys.stdout.write('========== [DONE] Command on #%d ==> %s (PWD=%s) ==========\n'
                         % (self.id, self.full_dir, self.full_dir))
        os.chdir(original_dir)
        return 0

    def sim_clean(self):
        """
        Clean-up the simulation directory. Leaving only input files and restart file there.

        :return: Return 0 if succeed, -1 if failed. If the simulation is running, clean cannot be performed. In such
        case, the method does nothing but returns 1.
        """
        return 0

    def sim_reset(self):
        """
        Clean-up the simulation directory. Leaving only input files in the simulation directory. Reset the current
        Simulation status to NOT STARTED.

        :return: Return 0 if succeed, -1 if failed.
        """
        return 0

    def sim_init(self):
        """
        Perform necessary initialization procedures in order to start the simulation. Note that this method will NOT
        start the simulation. It will only make the simulation ready to start when sim_start() is called.

        :return: Return 0 if succeed, -1 if failed. If the simulation is running/stopped/finished, the method does
        nothing but just return 1.
        """
        return 0

    def sim_finalize(self):
        """
        Finalize the simulation (e.g. perform data processing) after the simulation is finished.

        :return: Return 0 if succeed, -1 if failed. If the simulation is running/stopped or not yet started, the method
        does nothing but return 1.
        """
        return 0

    def sim_collect_recent_output_message(self, lines=20):
        """
        Collect the recent lines of output/error messages, generated by the simulation code.

        :return: Return the messages as a combined string if available. Otherwise return an empty string.
        """
        if self.config.has_option('Simulation', 'Output_file'):
            output_file = self.config.get('Simulation', 'Output_file')
            sys.stdout.write('========== Diagnose for #%d ==> %s ==========\n' % (self.id, self.full_dir))
            check_dir_name = self.full_dir
            original_dir = os.getcwd()
            os.chdir(check_dir_name)
            os.system('\ntail -%d %s' % (lines, output_file))
            restart_dir = sorted(glob.glob('restart*/'))
            for r_dir in restart_dir:
                os.chdir(r_dir)
                sys.stdout.write('========== Diagnose for restart ==> %s ==========\n' % r_dir)
                os.system('\ntail -%d %s' % (lines, output_file))
                os.chdir('..')
            os.chdir(original_dir)
        return str()



from setuptools import setup, find_packages

# name is the package name in pip, should be lower case and not conflict with existing packages
# packages are code source

setup(name='astrosimon',
      version='0.8.5',
      description='Simulation Monitor for computational astrophysics',
      url='https://github.com/maxwelltsai/SiMon',
      author='Maxwell Cai, Penny Qian',
      author_email='maxwellemail@gmail.com',
      license='BSD 2-Clause',
      packages=find_packages(),
      zip_safe=False,
      install_requires=['python-daemon', 'numpy'],
      entry_points={'console_scripts': ['simon = SiMon.simon:main'], },
      )



"""
Initial condition (IC) generator.
"""
import os
import sys
try:
    import configparser as cp  # Python 3 only
except ImportError:
    import ConfigParser as cp  # Python 2 only


class InitialConditionGenerator(object):

    config_file_template = '''# Per-simulation config file for SiMon
[Simulation]
# The name of the simulation code
Code_name = %s

# The file name of the initial condition input file (for stdin)
Input_file = %s

# The file name of the simulation log output (for stdout)
Output_file = %s

# The file name of the simulation error output (for stderr)
Error_file = %s

# The name of the file used to restart the simulation
Restart_file = %s

# The timestamp indicating the starting time of the simulation
Timestamp_started = %s

# The timestamp indicating the last time output files are updated
Timestamp_last_updated = %s

# The time (in second) beyond which a simulation is considered as stalled
Stall_time = %d

# The starting time
T_start = %f

# The termination time (i.e., t_max)
T_end = %f

# The process ID of the N-body code
PID = %d

# The priority (i.e. niceness) of the simulation (-20 to 19, lower are higher, same as UNIX systems)
Niceness = %d

# The shell command to start the simulation
Start_command: %s

# The shall command to restart the simulation
Restart_command: %s

# The shall command to stop the simulation
Stop_command: %s

# The maximum number of times a simulation will be restarted (a simulation is marked as ERROR when exceeding this limit)
Max_restarts: %d
    '''

    def __init__(self, conf_file):
        self.global_conf_file = conf_file  # the path of the global SiMon.conf
        self.config_file_per_sim = 'SiMon.conf'  # the per-simulation config file name
        self.config = None  # the parsed global config instance
        self.sim_data_dir = None  # simulation data dir
        self.max_restarts = 5  # maximum attempts a sim will be restarted, beyond which a sim is considered error
        self.stall_time = 6.e6  # Stall time

    def parse_config_file(self):
        print(self.global_conf_file)
        conf_fn = self.global_conf_file
        conf = cp.ConfigParser()
        if os.path.isfile(conf_fn):
            conf.read(conf_fn)
            self.config = conf
            if self.config.has_option('SiMon', 'Root_dir'):
                self.sim_data_dir = self.config.get('SiMon', 'Root_dir')
                # check whether the directory exist
                if not os.path.isdir(self.sim_data_dir):
                    print('Simulation data directory %s does not exist. Making the dir...' % self.sim_data_dir)
                    try:
                        os.makedirs(self.sim_data_dir)
                        print('Simulation data directory created successfully.')
                    except IOError as err:
                        print('Unable to create the simulation data directory: %s' % err)
                        print('Exiting...')
                        sys.exit(-1)
            else:
                print('Simulation root directory cannot be found in SiMon.conf. Existing...')
                sys.exit(-1)
            if self.config.has_option('SiMon', 'Max_restarts'):
                self.max_restarts = self.config.getint('SiMon', 'Max_restarts')
            if self.config.has_option('SiMon', 'Stall_time'):
                self.stall_time = self.config.getfloat('SiMon', 'Stall_time')

        else:
            print('Global config file %s cannot be found. Existing...' % self.global_conf_file)
            sys.exit(-1)

    def generate_simulation_ic(self, code_name, t_end, output_dir, start_cmd, input_file=None, output_file=None,
                               error_file=None, restart_file=None, t_stall=None, t_start=0, restart_cmd=None,
                               stop_cmd=None, niceness=0, max_restarts=None):
        """
        Generate the initial condition of a simulation and write it to the given directory.
        :param code_name: The name of the numerical code, e.g. DemoSimulation
        :param t_end: The termination time criterion
        :param output_dir: The directory in which the initial condition files and output files are contained
        :param start_cmd: The UNIX command to start the simulation
        :param input_file: The name of the input file (optional)
        :param output_file: The name of the output file (optional)
        :param error_file: The name of the error message file (optional)
        :param restart_file: The name of the restartable checkpoint file (optional)
        :param t_stall: How long (in sec) will a simulation be considered stalled since the last time the output file
                        is updated
        :param t_start: The starting model time (optional, default: 0)
        :param restart_cmd: The UNIX command to restart the simulation
        :param stop_cmd: The UNIX command to request the simulation code to stop the simulation (optional)
        :param niceness: The priority of the simulation, -20 to 19, lower are higher (optional, default: 0)
        :param max_restarts: The maximum number of attempts a simulation will be restarted, beyond which the simulation
                             is considered ERROR
        :return: return 0 if succeed, -1 if failed.
        """
        if max_restarts is None:
            max_restarts = self.max_restarts
        if t_stall is None:
            t_stall = self.stall_time
        if not os.path.isdir(os.path.join(self.sim_data_dir, output_dir)):
            print('Creating directory: %s' % os.path.join(self.sim_data_dir, output_dir))
            os.makedirs(os.path.join(self.sim_data_dir, output_dir))
        conf_file = open(os.path.join(self.sim_data_dir, output_dir, self.config_file_per_sim), 'w')
        conf_file.write(InitialConditionGenerator.config_file_template % (code_name,
                                                                          input_file,
                                                                          output_file,
                                                                          error_file,
                                                                          restart_file,
                                                                          0,  # timestamp started
                                                                          0,  # timestamp last modified
                                                                          t_stall,
                                                                          t_start,
                                                                          t_end,
                                                                          0,  # UNIX process ID (PID)
                                                                          niceness,
                                                                          start_cmd,
                                                                          restart_cmd,
                                                                          stop_cmd,
                                                                          max_restarts))  # max_restarts
        conf_file.close()





from ..simon import SiMon
import os
import unittest
import subprocess

# instance a simulation_task

class TestSimon(unittest.TestCase):
    def test_interactive_mode(self):
        # execute only if run as a script
        s = SiMon()

        # test interactive mode
        s.build_simulation_tree()
        s.print_sim_status_overview(0)

        # test choices without id input
        s.task_handler('l')

        with self.assertRaises(SystemExit):
            s.task_handler('q')

        '''
        with self.assertRaises(SystemExit) as cm:
            your_method()

        self.assertEqual(cm.exception.code, 1)
        '''

        # test choices requires id input
        self.selected_inst = 1  # test id

        # for choice in ['s', 'n', 'r', 'c', 'x', 'd', 'k', 'b', 'p']:
            # TODO: how to give an input here?
            # s.task_handler(choice)


    def test_daemon_mode(self):
        test_dir = os.path.join(os.path.dirname(__file__), 'test.sh')
        print('test dir', test_dir)
        subprocess.call([test_dir])
        # TODO: start and stop does not effect?